/*
 *****************************************************************************
 *
 * File:    get_login.h
 *
 * Author:  Damien S. Stuart
 *
 * Purpose: Header for getlogin.c 
 *
 * Copyright (C) 2008 Damien Stuart (dstuart@dstuart.org)
 *
 *  License (GNU Public License):
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *     USA
 *
 *****************************************************************************
*/
#ifndef GETLOGIN_H
#define GETLOGIN_H 1

/* Function prototypes
*/
char* getlogin (void);

#endif /* GETLOGIN_H */

/***EOF***/
